import numpy as np
import scipy.stats as sts

EPS = 1e-5


def linear_reg(x: np.ndarray, y: np.ndarray):
    """ Simple linear regression y[idx] = slope[idx] * x + intcp[idx]
    Args:
        x (np.ndarray): shape = (time,)
        y (np.ndarray): shape = (time,*number)
        x's dim0 must equal to y'dim0 !

    Returns:
        slope (np.ndarray): Slope, shape = [*number]
        intcpt (np.ndarray) : intercept , shape = [*number]
        corr (np.ndarray) : Simple correlation coefficient , shape = [*number]
        p_value (np.ndarray) : T test p value , shape = [*number]
    """
    # check shape
    Num0 = x.shape[0]
    Num1 = y.shape[0]
    # Initial Y shape
    org_sp_Y = y.shape
    # no equal, raise error
    if Num0 != Num1:
        raise (ValueError("x.shape[0] no equal to y.shape[0] , dim0 is wrong"))
    # reshape y to (time,number)
    y_rs = y.reshape(Num1, -1)
    # cal anomaly
    xa = x - x.mean(axis=0)
    ya = y_rs - y_rs.mean(axis=0)
    # cal std
    y_std = ya.std(axis=0)
    x_std = xa.std(axis=0)
    # cal covariance
    covar = ya.T @ xa / Num0
    # cal corr
    corr = covar / y_std / x_std
    # cal slope
    slope = corr * y_std / x_std
    # cal intcpt
    intcpt = y_rs.mean(axis=0) - slope * x.mean(axis=0)
    # cal t-valpue
    t = corr / (np.sqrt(1 - corr**2) + EPS) * np.sqrt(Num0 - 2)
    p_value = sts.t.sf(t, df=Num0 - 2)
    # transform t value
    pv_cp = np.copy(p_value)
    p_value[pv_cp >= 0.5] = (1 - p_value[pv_cp >= 0.5]) * 2
    p_value[pv_cp < 0.5] = (p_value[pv_cp < 0.5]) * 2
    # reshape
    slope, intcpt, corr, p_value = list(map(lambda inar: inar.reshape(org_sp_Y[1:]), [slope, intcpt, corr, p_value]))
    # return result
    return slope, intcpt, corr, p_value


def multi_linreg(x: np.ndarray, y: np.ndarray):
    """ multiple linear regression :: y[idx] = slope[idx,0] * x[0] + slope[idx, 1] * x[1] + ... + intcp[idx]

    Args:
        x (np.ndarray): shape = (time, N) ; N is Number of factors
        y (np.ndarray): shape = (time,*number)
        x's dim0 must equal to y'dim0 !
    Returns:
        slope (np.ndarray): Slope, shape = [N,*number]
        intcpt (np.ndarray) : intercept , shape = [*number]
        R (np.ndarray) : multiple correlation coefficient , shape = [*number]
        pv_all(np.ndarray) : F test p value , shape = [*number]
        pv_i(np.ndarray) : F test p value of every infact, shape = [N, *number]
    """
    # save shape
    Num0 = x.shape[0]
    Num1 = y.shape[0]
    # number of factors
    Numf = x.shape[1]
    # y initial shape
    org_sp_Y = y.shape
    # raise error
    if Num0 != Num1:
        raise (ValueError("x.shape[0] no equal to y.shape[0] , dim0 is wrong"))
    # reshape y
    y_rs = y.reshape(Num1, -1)
    # cal mean
    x_mean = x.mean(axis=0)
    y_mean = y_rs.mean(axis=0)
    # get anomaly
    xa = x - x_mean
    ya = y_rs - y_mean
    # cal covar
    covar_xx = xa.T @ xa / Num0
    covar_yx = xa.T @ ya / Num0
    # get slope
    slope = np.linalg.solve(covar_xx, covar_yx)  # (Numf , Num space)
    # get intcpt
    intcpt = y_mean - slope.T @ x_mean
    # get y_pre
    y_pre = xa @ slope
    # Total variance
    SSyy = Num0 * y_rs.var(axis=0)
    # Regression interpretation variance
    U = Num0 * y_pre.var(axis=0)
    # residual
    Q = SSyy - U
    # correlation coefficient
    R = np.sqrt(U / SSyy)  # (Num space)
    # F value of total regression equation
    F = U / Numf / (Q / (Num0 - Numf - 1))
    # F value of Infactor
    C_mx = np.diagonal(np.linalg.inv(covar_xx * Num0))
    Q_mx = slope**2 / C_mx[..., np.newaxis]  # (Numf,Num space)
    F_i = Q_mx / (Q / (Num0 - Numf - 1))  # (Numf,Num space)
    # p value
    pv_all = 1 - sts.f.cdf(F, Num0, Num0 - Numf - 1)  # (Num space)
    pv_i = 1 - sts.f.cdf(F_i, 1, Num0 - Numf - 1)  # (Numf,Num space)
    # reshape
    pv_all, R, intcpt = list(map(lambda inar: inar.reshape(org_sp_Y[1:]), [pv_all, R, intcpt]))
    slope, pv_i = list(map(lambda inar: inar.reshape([Numf, *org_sp_Y[1:]]), [slope, pv_i]))
    # return
    return slope, intcpt, R, pv_all, pv_i


def multi_corr(x: np.ndarray, y: np.ndarray):
    # save shape
    Num0 = x.shape[0]
    Num1 = y.shape[0]
    # number of factors
    Numf = x.shape[1]
    # y initial shape
    org_sp_Y = y.shape
    # raise error
    if Num0 != Num1:
        raise (ValueError("x.shape[0] no equal to y.shape[0] , dim0 is wrong"))
    # reshape y
    y_rs = y.reshape(Num1, -1)
    # Num space
    Nums = y_rs.shape[1]
    # cal mean
    x_mean = x.mean(axis=0)
    y_mean = y_rs.mean(axis=0)
    # get anomaly
    xa = x - x_mean
    ya = y_rs - y_mean
    # cal std
    x_std = xa.std(axis=0)
    y_std = ya.std(axis=0)
    # normalization
    x_norm = xa / x_std
    y_norm = ya / y_std
    # before correlation
    x_norm_b = np.repeat(x_norm[..., np.newaxis], axis=2, repeats=Nums)  # (t,Nf,Ns)
    y_norm_b = y_norm[:, np.newaxis, :]  # (t , 1 ,Ns)
    conc_xy = np.concatenate([x_norm_b, y_norm_b], axis=1)  # (t,Nf+1,Ns)
    corr_matrix = np.transpose(conc_xy, axes=[2, 1, 0]) @ np.transpose(conc_xy, axes=[2, 0, 1]) / Num0  # (Ns,Nf+1,Nf+1)
    corr_matrix = corr_matrix.reshape([*org_sp_Y[1:], Numf + 1, Numf + 1])
    return corr_matrix


def partial_corr(x: np.ndarray, y: np.ndarray, indx: int, mul_corr=None):
    """ calculate partial correlation

    Args:
        x (np.ndarray): shape = (time, N) ; N is Number of factors
        y (np.ndarray): shape = (time,*number)
        mul_corr (np.ndarray, optional): _description_. Defaults to None.
        idnx (np.ndarray) : correlation between y and x[:,indx], 
                Exclude the influence of other factors
        x's dim0 must equal to y'dim0 !
    Returns:
        np.ndarray: correlation between y and x[:,idx], Exclude the influence of other factors
    """
    if mul_corr is None:
        corr_matrix = multi_corr(x, y)
    Numf = corr_matrix.shape[1] - 1
    corr_matrix = corr_matrix.reshape((-1, Numf + 1, Numf + 1))
    Rsyy = np.linalg.det(corr_matrix[:, :-1, :-1])
    Rsii = np.linalg.det(np.delete(np.delete(corr_matrix, indx, 1), indx, 2))
    Rsyi = (-1)**(Numf + indx) * np.linalg.det(np.delete(np.delete(corr_matrix, indx, 1), Numf, 2))
    return -Rsyi / np.sqrt(Rsyy * Rsii)



import numpy as np
import xarray as xr
from .Util import _correct_type

EPS = 1e-5


class SVD():
    name = "SVD"
    """ SVD analysis of data
    """

    def __init__(self, data1: np.ndarray, data2: np.ndarray, complex=False):
        """ initiation of EOF
        Args:
            data1 and data2 (np.ndarray): shape (time, * space grid number)
        """
        # original data
        if complex:
            dtype_np = np.complex128
        else:
            dtype_np = np.float64
        self.dtype_np = dtype_np
        self._pc_std_save = 0
        # if isinstance(data1, xr.DataArray):
        #     data1 = np.array(data1, dtype=dtype_np)
        data1 = _correct_type(data1, dtype=dtype_np)
        # if isinstance(data2, xr.DataArray):
        #     data2 = np.array(data2, dtype=dtype_np)
        data2 = _correct_type(data2, dtype=dtype_np)
        self.origin_shape1 = data1.shape
        self.origin_shape2 = data2.shape
        # time length
        if data1.shape[0] != data2.shape[0]:
            raise ValueError("data1.time [%s] != data2.time [%s]" % (data1.shape[0], data2.shape[0]))
        self.tLen = data1.shape[0]
        # reshape (time, space)
        self.rsp_data1 = data1.reshape(self.tLen, -1)
        self.rsp_data2 = data2.reshape(self.tLen, -1)

    def _mask_nan(self, data):
        """ mask Nan or drop out the Nan
        """
        # get not Nan flag
        flag = np.isnan(data).sum(axis=0) == 0
        # save to self.flag
        self.flag = flag
        # get data without Nan
        data_nN = data[:, flag]  # (time, space_noNan)
        return data_nN, flag

    def solve(self):
        """ solve the SVD
        """
        # mask data
        data1_noNan, flag1 = self._mask_nan(self.rsp_data1)
        data2_noNan, flag2 = self._mask_nan(self.rsp_data2)
        self._data1_noNan, self._data2_noNan = data1_noNan, data2_noNan
        self.flag1, self.flag2 = flag1, flag2
        # get covariance
        Covan = data1_noNan.T @ data2_noNan
        # svd solver
        U_matrix, S, V_matrix = np.linalg.svd(Covan)
        # save
        self._U = U_matrix
        self._V = V_matrix
        self.eign = S

    def get_eign(self):
        """ return the eign of coviance"""
        return self.eign

    def get_varperc(self, npt):
        """ return percentile of variance eign"""
        var_perc = self.eign[:npt] / np.sum(self.eign)
        return var_perc

    def get_pc(self, npt, norm="std"):
        pc_left = self._U[:, :npt].T @ self._data1_noNan.T
        pc_right = self._V[:, :npt].T @ self._data2_noNan.T
        pc_left, pc_right = np.real(pc_left), np.real(pc_right)
        if norm == "std":
            self.pc_left_std = pc_left.std(axis=1)[:, np.newaxis]
            self.pc_right_std = pc_right.std(axis=1)[:, np.newaxis]
            pc_left /= self.pc_left_std
            pc_right /= self.pc_right_std
        else:
            self.pc_left_std = pc_left.std(axis=1)[:, np.newaxis]
            self.pc_right_std = pc_right.std(axis=1)[:, np.newaxis]
        self._pc_std_save = npt
        return pc_left, pc_right

    def get_pt(self, npt, norm="std"):
        #
        patterns_left = np.zeros((npt, *self.rsp_data1.shape[1:]), dtype=self.dtype_np)
        patterns_right = np.zeros((npt, *self.rsp_data2.shape[1:]), dtype=self.dtype_np)
        if norm == "std":
            if npt > self._pc_std_save:
                self.get_pc(npt=npt)
            else:
                pass
            patterns_left *= self.pc_left_std[:npt]
            patterns_right *= self.pc_right_std[:npt]
        else:
            pass
        # 
        patterns_left[:, self.flag1] = self._U[:, :npt].T
        patterns_left[:, np.logical_not(self.flag1)] = np.NAN
        patterns_left = patterns_left.reshape((npt, *self.origin_shape1[1:]))
        #

        patterns_right[:, self.flag2] = self._V[:npt]
        patterns_right[:, np.logical_not(self.flag2)] = np.NAN
        patterns_right = patterns_right.reshape((npt, *self.origin_shape2[1:]))

        return patterns_left, patterns_right



import xarray as xr
import os

current_dir = os.path.split(os.path.realpath(__file__))[0]


def load_sst():
    """ load example sst data
    """
    return xr.open_dataset(current_dir + "/data/example/HadISST_sst_5x5.nc")


def load_10mwind():
    """ load 10 m wind
    """

    return xr.open_dataset(current_dir + "/data/example/NCEP_wind10m_5x5.nc")



import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import xarray as xr


def convert_lon(lon: np.ndarray):
    """ Convert longitude to the range of [-180, 180]

    Args:
        lon (np.ndarray): lon in [0,360]

    Returns:
        lon (np.ndarray): lon in [-180 , 180]
    """
    return (lon + 180) % 360 - 180


def reverse_lon(lon: np.ndarray):
    """ Convert longitude to the range of [0, 360]

    Args:
        lon (np.ndarray): lon in [-180,180]

    Returns:
        lon (np.ndarray): lon in [0, 360]
    """
    lon1 = np.copy(lon)
    lon1[lon < 0] += 360
    return lon1


def wrap_lon_to_180(da, lon='lon'):
    '''
    This code copy from https://github.com/Yefee/xMCA
    Wrap longitude coordinates of DatArray to -180..179
    
    Parameters
    ----------
    da : DatArray
        object with longitude coordinates
    lon : string
        name of the longitude ('lon', 'longitude', ...)

    Returns
    -------
    wrapped : Dataset
        Another dataset array wrapped around.
    '''

    # wrap 0..359 to -180..179
    da = da.assign_coords(lon=(((da[lon] + 180) % 360) - 180))

    # sort the data
    return da.sortby(lon)


def sig_scatter(ax, x: np.ndarray, y: np.ndarray, p: np.ndarray, threshold: float = 0.05, **kwargs):
    """ significance plot 1D

    Args:
        ax (matplotlib.axes._subplots.AxesSubplot): axes to plot
        x (np.ndarray): 1D
        y (np.ndarray): 1D
        p (np.ndarray): 1D
        threshold (float, optional): threshold p value to plot. Defaults to 0.05.
    """
    # copy
    x1, y1, p1 = list(map(np.copy, [x, y, p]))
    x1[p1 >= threshold] = np.NAN
    y1[p1 >= threshold] = np.NAN
    # scatter
    sc = ax.scatter(x1, y1, **kwargs)
    return sc


def _correct_type0(data, dtype=np.float64):
    """ 

    Args:
        data (any): data to correct
        dtype (any, optional): data format. Defaults to np.float64.

    Returns:
        format data
    """
    if not isinstance(data, np.ndarray):
        if isinstance(data, xr.DataArray):
            data1 = data.to_numpy()
            # data = data.drop_duplicates()
            coords = data.coords
            dims = data.dims
        else:
            try:
                data1 = np.array(data)
                dims = None
                coords = None
            except:
                raise TypeError(f"Can't convert {type(data)} to {np.ndarray}'")
    else:
        data1 = data
        dims = None
        coords = None
    return data1,dims,coords


def _correct_type(data, dtype=np.float64):
    """ 

    Args:
        data (any): data to correct
        dtype (any, optional): data format. Defaults to np.float64.

    Returns:
        format data
    """
    if not isinstance(data, np.ndarray):
        if isinstance(data, xr.DataArray):
            data1 = data.to_numpy()
        else:
            try:
                data1 = np.array(data)
            except:
                raise TypeError(f"Can't convert {type(data)} to {np.ndarray}'")
    else:
        data1 = data
    return data1


def gradient_array(data: np.ndarray, dim: int, method: int = 0):
    """ gradient array in dim

    Args:
        data (np.ndarray): data need to calculate gradient
        dim (int) 
        method (int, optional): 0 is Central difference;
                        1 is front difference or half grid central difference;
                        . Defaults to 0.

    Returns:
        gradient array
    """
    data1 = _correct_type(data)
    data1_sp = data1.swapaxes(0, dim)
    if method == 0:
        grd_dat = data1_sp[2:] - data1_sp[:-2]
    elif method == 1:
        grd_dat = data1_sp[1:] - data1_sp[:-1]
    grd_dat = grd_dat.swapaxes(0, dim)
    return grd_dat


def gradient_da(da, dim, method=0, delta=None):

    if not isinstance(da, xr.DataArray):
        raise TypeError("'xr.DataArray' input is required, not the %s" % (type(da)))
    if not dim in list(da.coords.keys()):
        raise TypeError(f"DaArray don't have coords {dim} !")
    dim_coord = da[dim]
    if method == 0:
        dim_coord1 = dim_coord[1:]
        dim_coord2 = dim_coord[:-1]
    elif method == 1:
        dim_coord1 = dim_coord[2:]
        dim_coord2 = dim_coord[:-2]
    m_coord = (dim_coord1.to_numpy() + dim_coord2.to_numpy()) / 2
    # 没写完
    if delta is None:
        # delta_dim = {"lon":111e3,"lat":111e3,"level":5}
        if dim == "lat":
            delta = 1.11e5
        elif dim == "level":
            delta = 5
        elif dim == "lon":
            delta = 1.11e5 * np.cos(np.deg2rad(da["lat"]))
        else:
            delta = 1
    da1 = da.loc[{dim: dim_coord1}].assign_coords({dim: m_coord})
    da2 = da.loc[{dim: dim_coord2}].assign_coords({dim: m_coord})
    diff = (da1 - da2) / delta
    return diff



from .linger_cal import linear_reg, multi_linreg, multi_corr, partial_corr
import numpy as np
import xarray as xr
from scipy.spatial.distance import cdist
import scipy.stats as sts
from .Util import _correct_type,_correct_type0
from copy import copy, deepcopy

EPS = 1e-6


def gen_dataarray(data, coords):
    return xr.DataArray(data=data, coords=coords)


class LinReg:
    name = "LinReg"
    """
    Simple linear regression
        y[idx] = slope[idx] * x + intcp[idx]
    Attributes:
        slope (np.ndarray): Slope, shape = [*number]
        intcpt (np.ndarray) : intercept , shape = [*number]
        corr (np.ndarray) : Simple correlation coefficient , shape = [*number]
        p_value (np.ndarray) : T test p value , shape = [*number]
    """

    def __init__(self, x: np.ndarray, y: np.ndarray):
        """ Simple linear regression y[idx] = slope[idx] * x + intcp[idx]

        Args:
            x (np.ndarray): shape = (time,)
            y (np.ndarray): shape = (time,*number)
            x's dim0 must equal to y'dim0 !
        """
        # if isinstance(x, xr.DataArray):
        #     x = np.array(x)
        # if isinstance(y, xr.DataArray):
        #     y = np.array(y)
        x, xdims, xcoords = _correct_type0(x)
        y, ydims, ycoords = _correct_type0(y)
        # self.xdims, self.xcoords = xdims, xcoords
        # self.ydims, self.ycoords = ydims, ycoords

        # judge x.dim[0] and y.dim[0] length
        if x.shape[0] != y.shape[0]:
            raise ValueError(f"x.dim0 is {x.shape[0]} , y.dim0 is {y.shape[0]} ; not equal.")
        self.x = x
        self.y = y

        slope, intcpt, corr, p_value = linear_reg(x, y)

        # judge dataarray trans
        if ydims is not None and ycoords is not None:
            ycoords_m1 = {ydims[i]: ycoords[ydims[i]] for i in range(1, len(ydims))}
            slope, intcpt, corr, p_value = map(lambda data: gen_dataarray(data, ycoords_m1),
                                               [slope, intcpt, corr, p_value])
            # slope = gen_dataarray(slope, ycoords_m1)
        self.slope, self.intcpt, self.corr, self.p_value = \
            slope, intcpt, corr, p_value

    def _repr_html_(self):
        """
        show in jupyter
        """
        from .repr_html import res_repr_html
        return (res_repr_html(self))

    def __repr__(self) -> str:
        res = f"{self.name}, x.shape = {self.x.shape}, y.shape = {self.y.shape} "
        return res

    __str__ = __repr__


class M2mLinReg():
    name = "M2mLinReg"

    def __init__(self, x: np.ndarray, y: np.ndarray):
        """ calculate correlation of x[time,n_factor] and y[time,n_factor] and their p_value
        Args:
            x (np.ndarray): shape [time,n_factor]
            y (np.ndarray): shape [time,m_factor]

        Attribute:
            self.corr: [n_factor, m_factor]
            self.p_value [n_factor, m_factor]
        """
        # if isinstance(x, xr.DataArray):
        #     x = np.array(x)
        # if isinstance(y, xr.DataArray):
        #     y = np.array(y)
        x, xdims, xcoords = _correct_type0(x)
        y, ydims, ycoords = _correct_type0(y)
        if x.shape[0] != y.shape[0]:
            raise ValueError(f"x.dim0 is {x.shape[0]} , y.dim0 is {y.shape[0]} ; not equal.")
        Num0 = x.shape[0]
        x = x.reshape((x.shape[0], -1)).T
        y = y.reshape((y.shape[0], -1)).T
        corr = 1 - cdist(x, y, "correlation")
        self.corr = corr
        t = corr / (np.sqrt(1 - corr**2) + EPS) * np.sqrt(Num0 - 2)
        p_value = sts.t.sf(t, df=Num0 - 2)
        pv_cp = np.copy(p_value)
        p_value[pv_cp >= 0.5] = (1 - p_value[pv_cp >= 0.5]) * 2
        p_value[pv_cp < 0.5] = (p_value[pv_cp < 0.5]) * 2
        self.p_value = p_value

    def _repr_html_(self):
        """
        show in jupyter
        """
        from .repr_html import res_repr_html
        return (res_repr_html(self))


class MultLinReg:
    name = "MultLinReg"
    """
    multiple linear regression ::
        y[idx] = slope[idx,0] * x[0] + slope[idx, 1] * x[1] + ... + intcp[idx]
    Attribute:
        slope (np.ndarray): Slope, shape = [N,*number]
        intcpt (np.ndarray) : intercept , shape = [*number]
        R (np.ndarray) :
            multiple correlation coefficient , shape = [*number]
        pv_all(np.ndarray) : F test p value , shape = [*number]
        pv_i(np.ndarray) :
            F test p value of every infact, shape = [N, *number]
    """

    def __init__(self, x: np.ndarray, y: np.ndarray, cal_sim=True):
        """
        Args:
            x (np.ndarray): shape = (time, N) ; N is Number of factors
            y (np.ndarray): shape = (time,*number)
            cal_sim (Bool) : Whether to call function multi_linreg
            x's dim0 must equal to y'dim0 !
        """
        # if isinstance(x, xr.DataArray):
        #     x = np.array(x)
        # if isinstance(y, xr.DataArray):
        #     y = np.array(y)
        x, xdims, xcoords = _correct_type0(x)
        y, ydims, ycoords = _correct_type0(y)
        if x.shape[0] != y.shape[0]:
            raise ValueError(f"x.dim0 is {x.shape[0]} , y.dim0 is {y.shape[0]} ; not equal.")
        self.x = x
        self.y = y
        if cal_sim:
            self.slope, self.intcpt, self.R, self.pv_all, self.pv_i = multi_linreg(x, y)
        else:
            self.slope, self.intcpt, self.R, self.pv_all, self.pv_i = None, None, None, None
        
        self.multi_corr = None
        self.part_corr = {}

    def cal_corr(self):
        """ cal correlation between X_factors , y
        Returns:
            correlation : shape = [*y.shape[1:],Numf+1,Numf+1]
        """
        self.multi_corr = multi_corr(self.x, self.y)
        return self.multi_corr

    def cal_paritial_corr(self, idx: int):
        """calculate partial correlation 
        Args:
            idx (int): correlation between y and x[:,idx], Exclude the influence of other factors
        """
        self.part_corr[idx] = partial_corr(self.x, self.y, idx, mul_corr=self.multi_corr)
        return self.part_corr[idx]

    def _repr_html_(self):
        """
        show in jupyter
        """
        from .repr_html import res_repr_html
        return (res_repr_html(self))

    def __repr__(self) -> str:
        res = f"{self.name}, x.shape = {self.x.shape}, y.shape = {self.y.shape} "
        return res

    __str__ = __repr__



import sacpy as scp
import xarray as xr
import matplotlib.pyplot as plt
import sacpy.Map
import copy

sst = scp.load_sst()['sst']

coords = sst.coords
dim = sst.dims


coords1 = copy.deepcopy(coords)

del(coords1[dim[0]])

# print(coords.pop("time"))
print(coords1)

# newcoords = {dim[i]:coords[dim[i]] for i in range(len(dim))}
# newcoords.keys[0] = "time1"

# print(newcoords)
# print(dim)
# newc = coords

# print(type(coords))


try:
    from cartopy.mpl.geoaxes import GeoAxesSubplot
    import cartopy.crs as ccrs
    from cartopy.mpl.ticker import LongitudeFormatter, LatitudeFormatter
except:
    raise ImportError("Can't import cartopy, please check your envs")
from matplotlib.ticker import MultipleLocator
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import xarray as xr
from .Util import _correct_type
# set font to TNR
plt.rc('font', family='Times New Roman')


def get_levels(data, percentile: int, num_level: int, zero_sym: bool) -> np.ndarray:
    """ get levels of data

    Args:
        data (np.ndarray): data need to get levels
        percentile (int): max value or min value percentile
        num_level (int): colorbar levels
        zero_sym (bool): levels is zero symmetric or not

    Returns:
        np.ndarray: levels for contourf or contourf
    """
    #　set default values
    if percentile is None:
        percentile = 98
    if num_level is None:
        num_level = 13
    if zero_sym is None:
        zero_sym = True
    # get min and max
    _max = np.nanpercentile(data, percentile)
    _min = np.nanpercentile(data, 100 - percentile)
    if zero_sym is None:
        if _min*_max <0:
            zero_sym = True
        else:
            zero_sym = False
    # zero_sym :
    if zero_sym:
        max_real = np.max(np.abs([_max, _min]))
        levels = np.linspace(-max_real, max_real, num_level)
    else:
        levels = np.linspace(_min, _max, num_level)
    return levels


def _store_range(self, x, y):
    """ store data range for x and y of data

    Args:
        x (np.ndarray): x 
        y (np.ndarray): y
    """
    x= _correct_type(x)
    y= _correct_type(y)
    if x is not None:
        # save to xrange
        if not hasattr(self, "xrange"):
            self.xrange = [x]
        else:
            self.xrange.append(x)
    if y is not None:
        # save to yrange
        if not hasattr(self, "yrange"):
            self.yrange = [y]
        else:
            self.xrange.append(y)


def _get_extend(self):
    """ get self.xrange's xmin ,xmax and yrange's ymin and ymax
    Returns:
        xmin, xmax, ymin, ymax (float)
    """
    if hasattr(self, "xrange") and hasattr(self, "yrange"):
        xmin0, ymin0 = 360., 90.
        xmax0, ymax0 = -180, -90
        for xra, yra in zip(self.xrange, self.yrange):
            xmin1 = xra.min()
            ymin1 = yra.min()
            xmax1 = xra.max()
            ymax1 = yra.max()
            if xmin1 <= xmin0:
                xmin0 = xmin1
            if xmax1 >= xmax0:
                xmax0 = xmax1
            if ymin0 >= ymin1:
                ymin0 = ymin1
            if ymax0 <= ymax1:
                ymax0 = ymax1
        xmin, ymin, xmax, ymax = xmin0, ymin0, xmax0, ymax0
    else:
        xmin, ymin, xmax, ymax = None, None, None, None
    return xmin, xmax, ymin, ymax


def _get_extra_param(kwargs):
    """ get extra parameter for get levels

    Args:
        kwargs (dict)

    Returns:
        percentile, num_level, zero_sym (int,float,bool)
    """
    percentile = kwargs.get("percentile")
    if percentile is not None:
        del kwargs['percentile']
    num_level = kwargs.get("num_level")
    if num_level is not None:
        del kwargs['num_level']
    zero_sym = kwargs.get("zero_sym")
    if zero_sym is not None:
        del kwargs['zero_sym']
    return percentile, num_level, zero_sym


def _anal_args(args, clas="contourf"):
    """ analysis args for draw function

    Args:
        args (list)
        clas (str, optional): class of the function. Defaults to "contourf".

    Returns:
        np.ndarray: need value
    """
    if clas == "contourf":
        if len(args) == 1:
            levels0 = None
            Z = args[0]
            x, y = np.arange(Z.shape[0]), np.arange(Z.shape[1])
        elif len(args) == 2:
            levels0 = args[1]
            Z = args[0]
            x, y = np.arange(Z.shape[0]), np.arange(Z.shape[1])
        elif len(args) == 3:
            levels0 = None
            Z = args[2]
            x, y = args[0], args[1]
        elif len(args) == 4:
            levels0 = args[3]
            Z = args[2]
            x, y = args[0], args[1]
        else:
            raise ValueError("More parameters are transported")
        x = _correct_type(x)
        y = _correct_type(y)
        Z = _correct_type(Z)
        return Z, levels0, x, y
    elif clas == "pcolormesh":
        if len(args) == 1:
            Z = args[0]
            x, y = np.arange(Z.shape[0]), np.arange(Z.shape[1])
        elif len(args) == 2:
            raise ValueError("More parameters (2) are transported")
        elif len(args) == 3:
            Z = args[2]
            x, y = args[0], args[1]
        else:
            raise ValueError("More parameters are transported")
        x = _correct_type(x)
        y = _correct_type(y)
        Z = _correct_type(Z)
        return Z, x, y
    elif clas == "quiver":
        if len(args) == 1:
            raise ValueError("Need two arrays to depict wind")
        elif len(args) == 2:
            U, V = args[0], args[1]
            x, y = np.arange(U.shape[0]), np.arange(U.shape[1])
            C = None
        elif len(args) == 3:
            U, V = args[0], args[1]
            C = args[2]
            x, y = np.arange(U.shape[0]), np.arange(U.shape[1])
        elif len(args) == 4:
            x, y, U, V = args
            C = None
        elif len(args) == 5:
            x, y, U, V, C = args
        x = _correct_type(x)
        y = _correct_type(y)
        U = _correct_type(U)
        V = _correct_type(V)
        return x, y, U, V, C


def _contourf(self, *args, **kwargs):
    """ new contourf for GeoAxesSubplot

    Returns:
        mpl.contour.QuadContourSet
    """
    levels1 = kwargs.get("levels")
    extend = kwargs.get("extend")
    if extend is None:
        extend = "both"
        kwargs["extend"] = extend
    percentile, num_level, zero_sym = _get_extra_param(kwargs)
    Z, levels0, x, y = _anal_args(args)
    self._store_range(x, y)
    if levels0 is None and levels1 is None:
        levels = get_levels(Z, percentile, num_level, zero_sym)
        kwargs['levels'] = levels
    cmap = kwargs.get("cmap")
    if cmap is None:
        cmap = "RdBu_r"
        kwargs["cmap"] = cmap
    transform = kwargs.get("transform")
    if transform is None:
        transform = ccrs.PlateCarree()
        kwargs["transform"] = transform
    m = self.contourf(*args, **kwargs)
    return m


def _draw_ticks(self, extend, stepx=30, stepy=10, smallx=5, smally=2.5, bigx=10, bigy=10):
    """ draw map ticks

    Args:
        extend (list): extend of map [x1, x2, y1, y2]
        stepx (int, optional): big x step. Defaults to 30.
        stepy (int, optional): big y step. Defaults to 10.
        xsmall (int, optional): x small step. Defaults to 5.
        ysmall (int, optional): y small step. Defaults to 5.
        bigx/bigy (int, optional): Resolution in X and Y directions
    """
    [x1, x2, y1, y2] = extend
    # if stepx is None:
    #     stepx

    # autoly calculate stepx,stepy,smallx,smally,bigx,bigy
    # cling to 10 times
    xs = x1 // bigx * bigx if x1 % bigx == 0 else (x1 // bigx + 1) * bigx
    xe = x2 // bigx * bigx
    ys = y1 // bigy * bigy if y1 % bigy == 0 else (y1 // bigy + 1) * bigy
    ye = y2 // bigy * bigy
    # get xticks
    xticks = np.arange(xs, xe + 1, stepx)
    yticks = np.arange(ys, ye + 1, stepy)
    # set ticks
    self.set_xticks(xticks, crs=ccrs.PlateCarree())
    self.set_yticks(yticks, crs=ccrs.PlateCarree())
    self.yaxis.set_major_formatter(LatitudeFormatter())
    self.xaxis.set_major_formatter(LongitudeFormatter(zero_direction_label=True))
    self.xaxis.set_minor_locator(MultipleLocator(smallx))
    self.yaxis.set_minor_locator(MultipleLocator(smally))


def _initialize_map(self, same_size=True, coastlines=True, **kwargs):
    """ initial a map
    """
    # draw coastines
    if coastlines:
        self.coastlines()
    # Flatten map
    if same_size:
        self.set_aspect("auto")
    extend = self._get_extends()
    self.draw_ticks(extend, **kwargs)


def _contour(self, *args, **kwargs):
    """ new contour for GeoAxesSubplot

    Returns:
        mpl.contour.QuadContourSet
    """
    # get mpl.contour.QuadContourSet or not
    if not isinstance(args[0], (mpl.contour.QuadContourSet)):
        Z, levels0, x, y = _anal_args(args)
        self._store_range(x, y)
    # add transform
    transform = kwargs.get("transform")
    if transform is None:
        transform = ccrs.PlateCarree()
        kwargs['transform'] = transform
    n = self.contour(*args, **kwargs)
    return n


def _sig_ctrf(self, x, y, pvalue, thrshd=0.05, marker="..", color=None):
    """ Significance test dot

    Args:
        x (np.ndarray): x
        y (np.ndarray): y
        pvalue (np.ndarray): p value
        thrshd (float, optional): threshold of pvalue. Defaults to 0.05.
        marker (str, optional): mark of Significance test dot. Defaults to "..".
    """
    res = self.contourf(x,
                        y,
                        pvalue,
                        levels=[0, thrshd, 1],
                        zorder=1,
                        hatches=[marker, None],
                        colors="None",
                        transform=ccrs.PlateCarree())
    # set color
    if color is not None:
        for collection in res.collections:
            collection.set_linewidth(0.)
            collection.set_edgecolor(color)
    return res


def _pcolormesh(self, *args, **kwargs):
    """ new pcolormesh
    """
    Z, x, y = _anal_args(args, clas="pcolormesh")
    # store x ,y for ticks
    self._store_range(x, y)
    percentile, num_level, zero_sym = _get_extra_param(kwargs)
    vmin, vmax = kwargs.get("vmin"), kwargs.get("vmax")
    if vmin is None and vmax is None:
        levels = get_levels(Z, percentile, num_level, zero_sym)
        kwargs['vmin'] = levels[0]
        kwargs['vmax'] = levels[-1]
    cmap = kwargs.get("cmap")
    if cmap is None:
        cmap = "RdBu_r"
        kwargs["cmap"] = cmap
    transform = kwargs.get("transform")
    if transform is None:
        transform = ccrs.PlateCarree()
        kwargs["transform"] = transform
    m = self.pcolormesh(*args, **kwargs)
    return m


def _quiver(self, *args, **kwargs):
    """ new quiver
    """
    # add projcetion
    transform = kwargs.get("transform")
    if transform is None:
        transform = ccrs.PlateCarree()
        kwargs["transform"] = transform
    stepx = kwargs.get("stepx")
    if stepx is None:
        stepx = 1
    else:
        del kwargs['stepx']
    stepy = kwargs.get("stepy")
    if stepy is None:
        stepy = 1
    else:
        del kwargs['stepy']
    x, y, U, V, C = _anal_args(args, clas="quiver")
    if C is None:
        args1 = [x[::stepx], y[::stepy], U[::stepy, ::stepx], V[::stepy, ::stepx]]
    else:
        args1 = [x[::stepx], y[::stepy], U[::stepy, ::stepx], V[::stepy, ::stepx], C]
    m = self.quiver(*args1, **kwargs)
    return m


def _xr_splot(self, ax=None, label=0, kw1={}, kw2={}):
    if label == 0:
        lon = "lon"
        lat = "lat"
    else:
        lon = "longitude"
        lat = "latitude"
    if ax is None:
        ax = plt.axes(projection=ccrs.PlateCarree(central_longitude=self[lon].mean().item()))
    m = ax.scontourf(self[lon], self[lat], self.to_numpy(), **kw1)
    ax.init_map(**kw2)
    plt.colorbar(m)
    return m, ax


GeoAxesSubplot.scontourf = _contourf
GeoAxesSubplot.scontour = _contour
GeoAxesSubplot.draw_ticks = _draw_ticks
GeoAxesSubplot.init_map = _initialize_map
GeoAxesSubplot.sig_plot = _sig_ctrf
GeoAxesSubplot.spcolormesh = _pcolormesh
GeoAxesSubplot._store_range = _store_range
GeoAxesSubplot._get_extends = _get_extend
GeoAxesSubplot.squiver = _quiver
xr.DataArray.splot = _xr_splot


from .version import __version__
from .LinReg import LinReg, MultLinReg
from .XrTools import get_anom, spec_moth_dat, spec_moth_yrmean
from .Util import convert_lon, reverse_lon, sig_scatter
from .load_sst import load_sst,load_10mwind
from .EOF import EOF
from .linger_cal import linear_reg, multi_linreg, multi_corr, partial_corr
from .SigTest import one_mean_test, two_mean_test
from .SVD import SVD

__author__ = "Zilu Meng"
__email__ = "mzll1202@163.com"


import xarray as xr
import numpy as np
from scipy import signal


def _deTrend(array: np.ndarray):
    """ Remove linear trend
    Args:
        array (np.ndarray): shape = (time,*number)
            Array of linear trends that need to be removed
    Returns:
        array (np.ndarray): shape = (time,*number)
    """
    array = np.array(array)
    # Remove the missing measured value
    flag = np.isnan(array).sum(axis=0) != 0
    flags = np.repeat(flag[np.newaxis, ...], axis=0, repeats=array.shape[0])
    # Set the missing value to 0
    array[flags] = 0
    # Remove linear trend
    res = signal.detrend(array, axis=0)
    # Set the missing value to Nan
    res[flags] = np.NAN
    return res


def get_anom(DaArray: xr.DataArray, method=0):
    """ Get climate data anomaly
    Args:
        DaArray (xr.DataArray): shape = (time, *number) original Dataarray
        method (int, optional): method of getting anomaly.
            0 is Minus the multi-year average of the corresponding month
            1 is Remove the linear trend of the corresponding month
             Defaults to 0.

    Returns:
        anom (xr.DataArray): climate data anomaly
    """
    if not isinstance(DaArray, xr.DataArray):
        raise TypeError("'xr.DataArray' input is required, not the %s" % (type(DaArray)))
    if not "time" in list(DaArray.coords.keys()):
        raise TypeError("DaArray must have coords 'time' !")
    if method == 0:
        anom = DaArray.groupby("time.month") - DaArray.groupby("time.month").mean()
    if method == 1:
        anom = DaArray.groupby("time.month").map(_deTrend)

    return anom


def spec_moth_dat(DaArray: xr.DataArray, months: list or str):
    """ get specific month data 
    Args:
        DaArray (xr.DataArray): shape = (time, *number) original Dataarray
        months (list): get data from specific month

    Raises:
        ValueError: xr.DataArray' input is required, not the %s

    Returns:
        xr.DataArray: data in specific month
    """
    if not isinstance(DaArray, xr.DataArray):
        raise ValueError("'xr.DataArray' input is required, not the %s" % (type(DaArray)))
    if not "time" in list(DaArray.coords.keys()):
        raise ValueError("DaArray must have coords 'time' !")
    season_dict = {"DJF": [12, 1, 2], "MAM": [3, 4, 5], "JJA": [6, 7, 8], "SON": [9, 10, 11]}
    if isinstance(months, str):
        months = season_dict.get(months)
        if months is None:
            raise TypeError(f"months mush be in {season_dict.keys()}")
    time = DaArray.time
    time_label = time.dt.month == months[0]
    for i in range(len(months)):
        time_label = time_label | (time.dt.month == months[i])
    return DaArray[time_label]


def spec_moth_yrmean(DaArray: xr.DataArray, months: list or str):
    """ get specific month data and average them in each year

    Args:
        DaArray (xr.DataArray): shape = (time, *number) original Dataarray
        months (list): get data from specific month

    Returns:
        xr.DataArray: data in specific month and average them in each year
    """
    # get data in specifc month
    season_dict = {"DJF": [12, 1, 2], "MAM": [3, 4, 5], "JJA": [6, 7, 8], "SON": [9, 10, 11]}
    if isinstance(months, str):
        months = season_dict.get(months)
        if months is None:
            raise TypeError(f"months mush be in {season_dict.keys()}")
    smd = spec_moth_dat(DaArray, months)
    # get time
    time = smd.time
    # get month[0]
    month_start = time.dt.month[0].item()
    # get month[-1]
    month_end = time.dt.month[-1].item()
    # get cycle data (cycle is period in months)
    # ed_idx = -(len(months) - months.index(month_end) - 1)
    ed_idx = -((+months.index(month_end) + 1) % len(months))
    ed_idx = None if ed_idx == 0 else ed_idx
    st_idx = len(months) - months.index(month_start)
    if st_idx == len(months):
        st_idx = 0
    smd_cyc = smd[st_idx:ed_idx]
    # get year list
    year_list = smd_cyc[::len(months)].time.dt.year
    # sum
    smd_sum = 0
    for i in range(len(months)):
        smd_i = smd_cyc[i::len(months)]
        smd_i['time'] = year_list
        smd_sum = smd_sum + smd_i
    # get mean
    smd_m = smd_sum / len(months)
    # return
    return smd_m



import numpy as np
import scipy.stats as sts
import xarray as xr

EPS = 1e-5


class EOF:
    """ EOF analysis of data
    """
    name = "EOF"

    def __init__(self, data: np.ndarray, weights=None):
        """ initiation of EOF
        Args:
            data (np.ndarray): shape (time, * space grid number)
            weights : shape (* space grid number , 
                        or can be broadcast to space grid number)
        """
        # original data
        if isinstance(data, xr.DataArray):
            data = np.array(data)
        if weights is None:
            self.data = np.copy(data)
        else:
            self.data = np.copy(data) * weights
        # original data shape
        self.origin_shape = data.shape
        # time length
        self.tLen = data.shape[0]
        # reshape (time, space)
        self.rsp_data = data.reshape(self.tLen, -1)
        self.pc = None
        self.got_pc_num = 0

    def _mask_nan(self):
        """ mask Nan or drop out the Nan
        """
        # get data0
        data0 = self.rsp_data
        # get not Nan flag
        flag = np.isnan(data0).sum(axis=0) == 0
        # save to self.flag
        self.flag = flag
        # get data without Nan
        data_nN = data0[:, flag]  # (time, space_noNan)
        # save
        self.data_nN = data_nN

    # def _mask_extra_data(self,data):
    #     flag = np.isnan(data0).sum(axis=0) == 0

    def solve(self):
        """ solve the EOF
        """
        # mask data
        self._mask_nan()
        # solve maksed data
        data_nN = self.data_nN
        # get (time,spcae_noNan)
        dim0_len, dim1_len = data_nN.shape
        # time > space
        if dim0_len > dim1_len:  # time > space
            dim_min = dim1_len
            # get coviance (space_noNan,space_noNan)
            cov = data_nN.T @ data_nN
            # get eigenvalues and right eigenvectors
            eign, e_vector = np.linalg.eig(cov)  # (dim_min) ; (dim_min , dim_min) [i]&[:,i]
            # trans
            e_vector = e_vector.T  # [i]&[i,:]

        else:  # space > time
            # get coviance
            dim_min = dim0_len
            # get cov, (time,time)
            cov = data_nN @ data_nN.T
            # get eigenvalues and right eigenvectors
            eign, e_vector_s = np.linalg.eig(cov)  # (dim_min) ; (dim_min , dim_min) [i]&[:,i]
            # trans
            e_vector = (data_nN.T @ e_vector_s / np.sqrt(np.abs(eign))).T[:dim_min]

        # save
        self.e_vector = e_vector
        self.eign = eign
        self.dim_min = dim_min
        # get patterns
        patterns = np.zeros((dim_min, *self.rsp_data.shape[1:]))
        patterns[:, self.flag] = e_vector
        # refill Nan
        patterns[:, np.logical_not(self.flag)] = np.NAN
        # patterns = patterns.reshape((dim_min, *self.origin_shape[1:]))
        # save
        self.patterns = patterns

    def get_eign(self):
        """ get eign of each pattern
        """
        return self.eign

    def get_varperc(self, npt=None):
        """ return variance percential

        Args:
            npt (int, optional): n patterns to get. Defaults to None.

        Returns:
            variace percentile (np.ndarray): variance percentile (npt,)
        """
        if npt is None:
            npt = self.dim_min
        var_perc = self.eign[:npt] / np.sum(self.eign)
        return var_perc

    def get_pc(self, scaling="std", npt=None):
        """ get pc of eof analysis

        Args:
            scaling (str, optional): scale method. Defaults to "std".

        Returns:
            pc_re : pc of eof (pattern_num, time)
        """

        if npt is None:
            npt = self.dim_min
        pc = self.e_vector[:npt] @ self.data_nN.T  # (pattern_num, time)
        # self.pc = pc
        self.pc_std = pc[:npt].std(axis=1)[..., np.newaxis]
        if scaling == "std":
            pc_re = pc[:npt] / self.pc_std
        else:
            pc_re = pc
        self.pc = pc_re
        self.got_pc_num = npt
        return pc_re

    def get_pt(self, scaling="mstd", npt=None):
        """ get spatial patterns of EOF analysis

        Args:
            scaling (str, optional): sacling method. Defaults to "mstd".

        Returns:
            patterns : (pattern_num, * space grid number)
        """
        if npt is None:
            npt = self.dim_min
        if self.pc is None or self.got_pc_num < npt:
            self.get_pc(npt=npt)
        if scaling == "mstd":
            patterns = self.patterns[:npt] * self.pc_std[:npt]
        else:
            patterns = self.patterns[:npt]
        # reshape to original shape (pattern_num,*space)
        patterns = patterns.reshape((npt, *self.origin_shape[1:]))
        return patterns

    def projection(self, proj_field: np.ndarray, npt=None, scaling="std"):
        """ project new field to EOF spatial pattern

        Args:
            proj_field (np.ndarray): shape (time, *space grid number)
            scaling (str, optional): _description_. Defaults to "std".

        Returns:
            pc_proj : _description_
        """
        if npt is None:
            npt = self.dim_min
        proj_field_noNan = proj_field.reshape(proj_field.shape[0], -1)[:, self.flag]
        pc_proj = self.e_vector @ proj_field_noNan.T
        if self.pc is None or self.got_pc_num < npt:
            self.get_pc(npt)
        if scaling == "std":
            pc_proj = pc_proj / self.pc.std(axis=1)[..., np.newaxis]
        return pc_proj

    def pattern_corr(self, data: np.ndarray, npt=None):
        """  calculate pattern correlation of extra data and patterns

        Args:
            data (np.ndarray): shape (time, * space number)
            npt (int, optional): number of spatial patterns . Defaults to None.

        Returns:
            corr , p_value: corr and p_value shape (npt,data_time)
        """
        if npt is None:
            npt = self.dim_min
        # mask data
        data_noNan = data.reshape(data.shape[0], -1)[:, self.flag]
        # get need e_vector
        need_evctor = self.e_vector[:npt]
        # free degree
        N = need_evctor.shape[1] - 2
        # normalize data
        norm_evctor = (need_evctor - need_evctor.mean(axis=1)[..., np.newaxis]) / \
                      (need_evctor.std(axis=1)[..., np.newaxis])
        data_noNan_norm = (data_noNan - data_noNan.mean(axis=1)[..., np.newaxis]) / \
                          (data_noNan.std(axis=1)[..., np.newaxis])

        corr = norm_evctor @ data_noNan_norm.T / (N + 2)  # npatterns,data_time
        t_value = np.abs(corr / (EPS + np.sqrt(1 - corr**2)) * np.sqrt(N))
        p_value = sts.t.sf(t_value, df=N - 2) * 2
        return corr, p_value



from setuptools import setup, find_packages

with open("README.md","r") as fh:
      long_des1 = fh.read()
with open("CHANGELOG.md","r") as fh1:
      long_des2 = fh1.read()
long_des = long_des1 + "\n\n" + long_des2
# print(long_des)


print(find_packages())
setup(name="sacpy",
      version="0.0.15",
      url="https://github.com/ZiluM/sacpy",
      description="A repaid Statistical Analysis tool for Climate or Meteorology data.",
      py_modules=["sacpy"],
      include_package_data=True,
      keywords="meteorology data statistic climate",
      author="Zilu Meng",
      author_email="mzll1202@163.com",
      install_requires=["numpy", "scipy", "xarray"],
      package_data={'sacpy': ['data/example/HadISST_sst_5x5.nc','data/example/NCEP_wind10m_5x5.nc']},
      packages=find_packages(),
      long_description=long_des,
      long_description_content_type="text/markdown",
      license="MIT")






# This code is from  Lifei Lin (Sun Yat-sen University) 
import uuid
from collections import OrderedDict
from functools import lru_cache, partial
from html import escape
from importlib.resources import read_binary
import xarray as xr
from xarray.core.formatting import inline_variable_array_repr, short_data_repr, format_array_flat
from xarray.core.options import _get_boolean_with_default
import numpy as np

STATIC_FILES = (
    ("xarray.static.html", "icons-svg-inline.html"),
    ("xarray.static.css", "style.css"),
)

def res_repr_html(arr):
    arr_name = arr.name
    header_components = [
        f"<h3>{arr_name}</h3>",
        # f"<div class='xr-obj-type'>{obj_type}</div>",
        # f"<div class='xr-array-name'>{arr_name}</div>",
    ]
    res_dict = arr.__dict__.copy()
    sections = [attribute_section(res_dict)]
    return _obj_repr(arr, header_components, sections)

@lru_cache(None)
def _load_static_files():
    """Lazily load the resource files into memory the first time they are needed"""
    return [
        read_binary(package, resource).decode("utf-8")
        for package, resource in STATIC_FILES
    ]

def _obj_repr(obj, header_components, sections):
    """Return HTML repr of an xarray object.

    If CSS is not injected (untrusted notebook), fallback to the plain text repr.

    """
    header = f"<div class='xr-header'>{''.join(h for h in header_components)}</div>"
    sections = "".join(f"<li class='xr-section-item'>{s}</li>" for s in sections)

    icons_svg, css_style = _load_static_files()
    return (
        "<div>"
        f"{icons_svg}<style>{css_style}</style>"
        f"<pre class='xr-text-repr-fallback'>{escape(repr(obj))}</pre>"
        "<div class='xr-wrap' hidden>"
        f"{header}"
        f"<ul class='xr-sections'>{sections}</ul>"
        "</div>"
        "</div>"
    )

def collapsible_section(
    name, inline_details="", details="", n_items=None, enabled=True, collapsed=False
):
    # "unique" id to expand/collapse the section
    data_id = "section-" + str(uuid.uuid4())

    has_items = n_items is not None and n_items
    n_items_span = "" if n_items is None else f" <span>({n_items})</span>"
    enabled = "" if enabled and has_items else "disabled"
    collapsed = "" if collapsed or not has_items else "checked"
    tip = " title='Expand/collapse section'" if enabled else ""

    return (
        f"<input id='{data_id}' class='xr-section-summary-in' "
        f"type='checkbox' {enabled} {collapsed}>"
        f"<label for='{data_id}' class='xr-section-summary' {tip}>"
        f"{name}:{n_items_span}</label>"
        f"<div class='xr-section-inline-details'>{inline_details}</div>"
        f"<div class='xr-section-details'>{details}</div>"
    )

def _mapping_section(
    mapping, name, details_func, max_items_collapse, expand_option_name, enabled=True
):
    n_items = len(mapping)
    expanded = _get_boolean_with_default(
        expand_option_name, n_items < max_items_collapse
    )
    collapsed = not expanded

    return collapsible_section(
        name,
        details=details_func(mapping),
        n_items=n_items,
        enabled=enabled,
        collapsed=collapsed,
    )

def short_data_repr_html(array):
    """Format "data" for DataArray and Variable."""
    internal_data = getattr(array, "variable", array)._data
    if hasattr(internal_data, "_repr_html_"):
        return internal_data._repr_html_()
    text = escape(short_data_repr(array))
    return f"<pre>{text}</pre>"

def _icon(icon_name):
    # icon_name should be defined in xarray/static/html/icon-svg-inline.html
    return (
        "<svg class='icon xr-{0}'>"
        "<use xlink:href='#{0}'>"
        "</use>"
        "</svg>".format(icon_name)
    )

def summarize_attrs(name, var, is_index=False, dtype=None, preview=None):
    if hasattr(var,"variable"):
      variable = var.variable
      dims_str = f"({', '.join(dim for dim in var.dims)})"
      dtype = dtype or escape(str(variable.dtype))
      data_repr = var._repr_html_()
    else:
      isDict = type(var) == dict
      isStr = type(var) == str
      isList = type(var) == list
      variable = xr.DataArray(var)
      dims = np.shape(variable)
      lens = len(dims)
      if(lens == 0): dims_str = ""
      elif(lens == 1): dims_str = dims[0]
      else: dims_str = dims
      dtype = dtype or escape(str(variable.dtype))
      if(isDict): dtype = "dict"
      elif(isStr): dtype = "str"
      data_repr = f"<p></p><ul class='xr-var-list'>{var}</ul><p></p>"

    disabled = "disabled"

      
    cssclass_idx = " class='xr-has-index'" if is_index else ""
    name = escape(str(name))
    # dtype = "dtype_test"

    # "unique" ids required to expand/collapse subsections
    attrs_id = "attrs-" + str(uuid.uuid4())
    data_id = "data-" + str(uuid.uuid4())

    preview = preview or escape(format_array_flat(variable, 35))

    attrs_icon = _icon("icon-file-text2")
    data_icon = _icon("icon-database")
    return (
        f"<div class='xr-var-name'><span{cssclass_idx}>{name}</span></div>"
        f"<div class='xr-var-dims'>{dtype}</div>"
        f"<div class='xr-var-dtype'>{dims_str}</div>"
        f"<div class='xr-var-preview xr-preview'>{preview}</div>"
        f"<input id='{attrs_id}' class='xr-var-attrs-in' "
        f"type='checkbox' {disabled}>"
        f"<label for='{attrs_id}' title='Show/Hide attributes'>"
        f"{attrs_icon}</label>"
        f"<input id='{data_id}' class='xr-var-data-in' type='checkbox'>"
        f"<label for='{data_id}' title='Show/Hide data repr'>"
        f"{data_icon}</label>"
        # f"<div class='xr-var-attrs'>{attrs_ul}</div>"
        f"<div class='xr-var-data'>{data_repr}</div>"
    )

def summarize_vars(dict):

    vars_li = "".join(
        f"<li class='xr-var-item'>{summarize_attrs(k, v)}</li>"
        for k, v in zip(dict, dict.values())
    )

    return f"<ul class='xr-var-list'>{vars_li}</ul>"

attribute_section = partial(
    _mapping_section,
    name="Attributes",
    details_func=summarize_vars,
    max_items_collapse=15,
    expand_option_name="display_expand_data_vars",
)


import numpy as np
import scipy.stats as sts
import xarray as xr


def one_mean_test(data: np.ndarray, expected_mean=None, *param) -> np.ndarray:
    """ one sample t test
    Args:
        data (np.ndarray): test data
        expected_mean (np.ndarray, optional): expected mean. Defaults to None.
        *param: parameter transports to sts.ttest_1samp

    Raises:
        TypeError: data's type should be np.ndarray
        ValueError: expected_mean shape can match data

    Returns:
        origin_mean,pvalue : np.ndarray: origin_mean and res.value
    """
    if isinstance(data,(xr.DataArray,np.ndarray)):
        data = np.array(data).copy()
    else:
        raise TypeError("Type should be np.ndarray or xr.Dataarray, rather than %s" % type(data))
    if expected_mean is None:
        expected_mean = np.zeros(data.shape[1:])
    else:
        if expected_mean.shape != data.shape[1:]:
            raise ValueError("expected_mean shape can match data")
    origin_mean = data.mean(axis=0)
    res = sts.ttest_1samp(data, expected_mean, axis=0, *param)
    return origin_mean, res.pvalue


def two_mean_test(data1: np.ndarray, data2: np.ndarray, *param) -> np.ndarray:
    """two samples t test

    Args:
        data1 (np.ndarray): test data1
        data2 (np.ndarray): test data2

    Raises:
        TypeError: Type should be np.ndarray

    Returns:
        mean_diff, pvalue np.ndarray: origin_mean difference and res.value
    """
    if isinstance(data1,(xr.DataArray,np.ndarray)) and isinstance(data2,(xr.DataArray,np.ndarray)):
        data1 = np.array(data1).copy()
        data2 = np.array(data2).copy()
    else:
        raise TypeError("Type should be np.ndarray or xr.Dataarray, rather than %s" % type(data1))
    mean_diff = data1.mean(axis=0) - data2.mean(axis=0)
    res = sts.ttest_ind(data1, data2, axis=0,*param)
    return mean_diff, res.pvalue
